# flcheck --
#
# This file implements package flcheck, a fusion logic modelchecker 
# using the CUDD BDD package
#
# Copyright (C) 2009, 2010  Antonio Cau

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

#
#

::bdd_tcl::load_file policy_library.tcl

#########################################################################


proc subjects {} {
    return [concat [users] [roles] ]
}

proc users {} {
    return [list  AC HJ ]
}


proc objects {} {
    return [list RM SERVER]
}

proc roles {} {
    return [list ADMIN USER]
}


#proc actions {} {
#    return [list  CREATE [actions_activate] [actions_deactivate] ]
#}


proc s_acau {} {
    return "[and AC {not HJ}]"  
}

proc s_helge {} {
    return "[and HJ {not AC}]" 
}

proc actions { object } {
    if { $object == "RM" } { 
        return [actions_activate] 
    } else {
      if {$object == "SERVER"} {
        return "CREATE"
      } else {
         return {}
      }
    }
}

proc actions_activate {} {
    set actions {}
    foreach a [roles] {
      lappend actions "ACTIVATE$a"
    }
     return $actions 
}

proc actions_deactivate {} {
    set actions {}
    foreach a [roles] {
      lappend actions "DEACTIVATE$a"
    }
     return $actions 
}

proc action_activate { role } {
   return "ACTIVATE$role"
}

proc action_deactivate { role } {
   return "DEACTIVATE$role"
}

proc inrole { user role } {
    set x "_"
    set y "ROLE$x$role"
    return "$user$x$y"
}


proc autho_plus { s o a } {
   set x "_"
   return "AUTHOP$x$s$x$o$x$a"
} 

proc autho_minus { s o a } {
   set x "_"
   return "AUTHON$x$s$x$o$x$a"
} 

proc autho_derived { s o a } {
   set x "_"
   return "AUTHO$x$s$x$o$x$a"
} 

proc ill { user } {
  set y "ILL"
  set x "_"
  return "$y$x$user"
}

proc not_ill { user } {
  set y "ILL"
  set x "_"
    return "{not $y$x$user}"
}


proc collect_set {} {
    global consequences
    array unset consequences
    foreach subject [subjects] {
     foreach object [objects] {
      foreach action [actions $object] {
          array set consequences [list [autho_plus $subject $object $action] 1]
          array set consequences [list [autho_minus $subject $object $action] 1]
      }
     }
    }
}



#proc rule_create_1 {} {
#    global consequences
#    set c [autho_plus ADMIN SERVER CREATE]
#    array set consequences [list $c 0]
#    return "[p_rule_1_true $c]"
#}

#proc rule_create_2 {} {
#    global consequences
#    set c [autho_minus USER SERVER CREATE]
#    array set consequences [list $c 0]
#    return "[p_rule_1_true $c]"
#}

proc rule_conflict_resolution {} {
  set l_rules {}
    foreach subject [subjects] {
     foreach object [objects] {
      foreach action [actions $object] {
          set x1 "not { [autho_minus $subject $object $action]} "
          set x [and [autho_plus $subject $object $action] $x1]
            set y [autho_derived $subject $object $action]
            set rule [strong_p_rule_s_1_0 $y $x]
            lappend l_rules $rule
      }
     }
    }
    set z [l_and $l_rules]
    return $z
}


#proc rule_activate_1_n {} {
#    global consequences
#    set c [autho_minus AC RM [action_activate ADMIN]]
#    array set consequences [list $c 0]
#    return [always [strong_p_rule_s_1_0 true $c]]
#}

#proc conflict_res {} {
#    set x1 "not [autho_minus AC RM [action_activate ADMIN]]"
#    set x [and [autho_plus AC RM [action_activate ADMIN]] $x1]
#    set y [autho_derived AC RM [action_activate ADMIN]]
#    set z [strong_p_rule_s_1_0 $x $y]
#    return [always $z]
#}

#proc no_conflict {} {
#    set x1 [autho_minus AC RM [action_activate ADMIN]]
#    set x2 [autho_plus AC RM [action_activate ADMIN]] 
#    set y  "not [and $x1 $x2]"       
#    return [always $y]
#}

#proc test_5 {} {
#    set x [list [rule_activate_1_n] [rule_activate_1] [conflict_res] ]
#    return [l_and $x]
#}

#proc test_6 {} {
#    set x [list [rule_activate_1_n] [rule_activate_1]  ]
#    set y [l_and $x]
#    return [implies $y [no_conflict]]
#}

#proc test_7 {} {
#    return [implies [test_5] [no_conflict]]
#}

proc rule_activate_1 {} {
    global consequences
    set c [autho_plus AC RM [action_activate ADMIN]]
    array set consequences [list $c 0]
    #puts "$c : $consequences($c)"
    return [strong_p_rule_s_1_0 true $c]
}

proc rule_activate_2 {} {
    global consequences
    set c [autho_plus HJ RM [action_activate USER]]
    array set consequences [list $c 0]
    return [strong_p_rule_s_1_0 true $c]
}

proc rule_activate_3 {} {
    global consequences
    set c [autho_plus HJ RM [action_activate ADMIN]]
    array set consequences [list $c 0]
    return [strong_p_rule_s_1_0 [ill AC] $c]
}


proc rule_activate_4 {} {
   global consequences
   set l_rules {}
    foreach user [users] {
      foreach action [actions "RM"] {
          set x [ill $user]
          set c [autho_minus $user "RM" $action]
          array set consequences [list $c 0]
          #puts "$c : $consequences($c)"
          set rule [strong_p_rule_s_1_0 $x $c]
          lappend l_rules $rule
      }
    }
    set z [l_and $l_rules]
    return $z
}




#proc rule_in_role {} {
#    global consequences
#    set l_rules {}
#    foreach user [users] {
#     foreach role [roles] {
#      foreach object [objects] {
#       foreach action [actions $object] {
#            set x [and [inrole $user $role] [autho_plus $role $object $action]]
#            set c [autho_plus $user $object $action]
#            array set consequences [list $c 0]
#            set rule [strong_p_rule_s_1_0 $x $c]
#            lappend l_rules $rule
#       }
#      }
#     }
#    }
#    set z [l_and $l_rules]
#    return $z
#}

proc default_rules {} {
   global consequences
    set l_rules {}
    foreach c [array names consequences] {
    #puts "$c : $consequences($c)"
        if { $consequences($c) == 1 } {
            set x [false]
            set rule [strong_p_rule_s_1_0 $x $c]
            lappend l_rules $rule
        }
    }
    if { [llength $l_rules] == 0 } {
      return "true"
    } else {
      set z [l_and $l_rules]
      return $z
    }
}




#proc constraint_1 {} {
#set l_rules {}
#    foreach user [users] {
#     foreach role [roles] {
#        set x1 [autho_derived $user "RM" "ACTIVATE_$role"]
#        set x "not $x1"
#        set y "not [inrole $user $role]"
#          set rule [implies $x $y]
#        lappend l_rules $rule
#     }
#    }
#    set z [l_and $l_rules]
#return $z
#}


#proc test_2 {} {
#  set y "not true"
#  set r1  [strong_p_rule_s_1_0 $y A_A]
#  set r2 [strong_p_rule_s_1_0 $y B_B]
#    return [always [and $r1 $r2] ]
#}

#proc test_3 {} {
#  set y "not true"
#    set r1  [list [strong_p_rule_s_1_0 $y A_A] [strong_p_rule_s_1_0 $y B_B] ]
#    return [always [l_and $r1 ] ]
#}

#proc test_4 {} {
#  set y "not true"
#    set r1  [list [strong_p_rule_s_1_0 $y A_A] [strong_p_rule_s_1_0 $y B_B] ]
#    return [always [l_and $r1 ] ]
#}


# test for validity
# not valid
proc test_1 {} { 
    return "[implies [always [model_1]] [property_1]]"
}

# test for validity
# not valid
proc test_2 {} { 
    return "[implies [always [model_2]] [property_1]]"
}

# test for satisfiability
# satisfiable
proc test_3 {} {
    return "[and [always [model_2]] [property_2]]"
}

# test for validity
# valid
proc test_4 {} { 
    return "[implies [always [model_2]] [property_3]]"
}

# test for validity
# not valid
proc test_5 {} { 
    return "[implies [always [model_2]] [property_4]]"
}

proc not_both_ill {} {
  set x "[and [ill HJ] [ill AC]]"
  set y "not $x"
  return "$y"
}

# test for validity
# valid
proc test_6 {} { 
    return "[implies [always [and [model_2] [not_both_ill]]] [property_4]]"
}


proc model_1 {} {
    collect_set
    set x [list \
            [rule_activate_1] \
            [rule_activate_2] \
            [rule_activate_3] \
            [rule_activate_4] \
          ]
    set y [default_rules]
    return [and [l_and $x] $y]
}


proc model_2 {} {
    collect_set
    set x [list \
            [rule_activate_1] \
            [rule_activate_2] \
            [rule_activate_3] \
            [rule_activate_4] \
            [rule_conflict_resolution] \
          ]
    set y [default_rules]
    return [and [l_and $x] $y]
}

proc property_1 {} {
    foreach subject [subjects] {
     foreach object [objects] {
      foreach action [actions $object] {
          set x "[autho_minus $subject $object $action]"
          set y "[autho_plus $subject $object $action]"
          set rule " not { [and $x $y] } "
          lappend l_rules $rule
      }
     }
    }
    set z [l_and $l_rules]
    return [always $z]
}

proc property_2 {} {
    return "[sometimes [autho_derived HJ RM [action_activate ADMIN]]]"
 
}


proc property_3 {} {
    set x [autho_derived AC RM [action_activate ADMIN]]
    set y [autho_derived HJ RM [action_activate ADMIN]]
    set z [and $x $y]
    set z1 "not { $z }"
    return "[always $z1]"
}

proc property_4 {} {
    set x [autho_derived AC RM [action_activate ADMIN]]
    set y [autho_derived HJ RM [action_activate ADMIN]]
    set z "{$x} or {$y}"
    return "[always $z]"
}

###

proc subjects_1 {} {
    return [concat [users_1] [roles_1] ]
}

proc users_1 {} {
   return [list  AC HJ ]
}

proc objects_1 {} {
    return [list RM]
}


proc roles_1 {} {
     return [list ADMIN]
}

proc collect_set_1 {} {
    global consequences
    array unset consequences
    foreach subject [users_1] {
     foreach object [objects_1] {
      foreach action [actions $object] {
          array set consequences [list [autho_plus $subject $object $action] 1]
          array set consequences [list [autho_minus $subject $object $action] 1]
      }
     }
    }
}

proc rule_conflict_resolution_1 {} {
  set l_rules {}
    foreach subject [users_1] {
     foreach object [objects_1] {
      foreach action [actions $object] {
          set x1 "not { [autho_minus $subject $object $action]} "
          set x [and [autho_plus $subject $object $action] $x1]
            set y [autho_derived $subject $object $action]
            set rule [strong_p_rule_s_1_0 $y $x]
            lappend l_rules $rule
      }
     }
    }
    set z [l_and $l_rules]
    return $z
}

proc rule_activate_1_1 {} {
    global consequences
    set c [autho_plus AC RM [action_activate ADMIN]]
    array set consequences [list $c 0]
    #puts "$c : $consequences($c)"
    return [strong_p_rule_s_1_0 true $c]
}

proc rule_activate_2_1 {} {
    global consequences
    set c [autho_plus HJ RM [action_activate USER]]
    array set consequences [list $c 0]
    return [strong_p_rule_s_1_0 true $c]
}


proc rule_activate_3_1 {} {
    global consequences
    set c [autho_plus HJ RM [action_activate ADMIN]]
    array set consequences [list $c 0]
    return [strong_p_rule_s_1_0 [ill AC] $c]
}

proc rule_activate_4_1 {} {
   global consequences
   set l_rules {}
    foreach user [users_1] {
      foreach action [actions "RM"] {
          set x [ill $user]
          set c [autho_minus $user "RM" $action]
          array set consequences [list $c 0]
          #puts "$c : $consequences($c)"
          set rule [strong_p_rule_s_1_0 $x $c]
          lappend l_rules $rule
      }
    }
    set z [l_and $l_rules]
    return $z
}

proc default_rules_1 {} {
   global consequences
    set l_rules {}
    foreach c [array names consequences] {
    #puts "$c : $consequences($c)"
        if { $consequences($c) == 1 } {
            set x [false]
            set rule [strong_p_rule_s_1_0 $x $c]
            lappend l_rules $rule
        }
    }
    if { [llength $l_rules] == 0 } {
      return "true"
    } else {
      set z [l_and $l_rules]
      return $z
    }
}


proc input_0 {} {
    return "[and [len 0] [both_not_ill]]"
}

proc both_not_ill {} {
  set x "[and [not_ill HJ] [not_ill AC]]"
  return "$x"
}

proc test_7 {} {
    return "[and [always [model_1_1]] [input_0]]"
}


proc input_0_enf {} {
    return [list [tl_and [list  [not_ill HJ] [not_ill AC] ] ] ]
}

proc test_7_enf {} {
    global INPUT
    set INPUT [input_0_enf]
    return "[always [model_1_enf] ]"
}

proc test_7_i {} {
    return "[always [model_1_enf] ]"
}

#proc input_1 {} {
#    return "[and [len 1] [always [both_not_ill]]]"
#}

proc input_1 {} {
    return "< test { [and [not_ill AC] [ill HJ]] } ; step { true} > { [and [not_ill AC] [not_ill HJ]] }"
}

proc test_8 {} {
    return "[and [always [model_1_1]] [input_1]]"
}

proc input_1_enf {} {
    return [list [tl_and [list [not_ill HJ] [not_ill AC] ]] [tl_and [list [not_ill HJ] [ill AC] ] ] ]
}

proc test_8_enf {} {
    global INPUT
    set INPUT [input_1_enf]
    return "[always [model_1_enf]]"
}


#proc input_2 {} {
#    return "[and [len 2] [always [both_not_ill]]]"
#}

proc input_2 {} {
    return "< test { [and [not_ill AC] [ill HJ]] } ; step { true} ; test { [and [ill AC] [not_ill HJ]] } ; step { true} > { [and [ill AC] [ill HJ]] }"
}

proc test_9 {} {
    return "[and [always [model_1_1]] [input_2]]"
}

proc input_2_enf {} {
    return [list [tl_and [list [not_ill HJ] [not_ill AC] ]] [tl_and [list [not_ill HJ] [ill AC] ] ] [tl_and [list [ill HJ] [ill AC] ] ] [tl_and [list [ill HJ] [not_ill AC] ] ] ]
}

proc test_9_enf {} {
    global INPUT
    set INPUT [input_2_enf]
    return "[always [model_1_enf]]"
}


proc model_1_1 {} {
    collect_set_1
    set x [list \
            [rule_activate_1_1] \
            [rule_activate_2_1] \
            [rule_activate_3_1] \
            [rule_activate_4_1] \
            [rule_conflict_resolution_1] \
          ]
    set y [default_rules_1]
    return [and [l_and $x] $y]
}

proc model_1_enf {} {
    collect_set_1
    set x [list \
            [rule_activate_1_1] \
            [rule_activate_2_1] \
            [rule_activate_3_1] \
            [rule_activate_4_1] \
            [rule_conflict_resolution_1] \
          ]
    set y [default_rules_1]
    set y [and [equiv [ill HJ] [ill HJ]] $y]
    return [and [l_and $x] $y]
}

# keys example

proc keys {} {
    set p_1 "[finite_e];test{[and KA KB ]}; [skip_e];test{true}"  
    set p_2 "[finite_e];test{true};[skip_e] ; test{[and KA  KB]}"
    set p_3 "[finite_e];test{KA};step{true} ; test{KB}"   
    set p_4 "[finite_e];test{KB};step{true} ; test{KA}" 
    set p "$p_1 or $p_2 or $p_3 or $p_4"
    set x [p_rule_1 $p A]
    return $x
}

proc keys_not {} {
    set p_1 "[finite_e];test{not KB};step{true};test{not KB}"
    set p_2 "[finite_e];test{not KA};step{true};test{not KA}"
    set p_3 "test{true}"
    set p "$p_1  or $p_2 or $p_3"
    set x [p_rule_1 $p "{not A}"]
    return $x
}
proc not_key { key } {
    return "{not $key}"
}

proc input_keys_enf {} {
   return [list [tl_and [list KA KB ] ] [tl_and [list [not_key KA] KB ] ] \
                [tl_and [list KA [not_key KB] ] ] \
                [tl_and [list KA [not_key KB] ] ] \
                [tl_and [list [not_key KA] KB ] ] \
                [tl_and [list [not_key KA] [not_key KB] ] ] \
                [tl_and [list KA KB ] ] ]
}



proc test_keys_enf {} {
    global INPUT
    set INPUT [input_keys_enf]
    return "[and [keys_not] [keys]]"
}
